#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import os
from lastfm import LastFM
from threading_util import ThreadPool, Future
from threading import Thread, Lock

from util import Util

import logging
log = logging.getLogger("TagPlayer")

class Builder:
	'''
	Builds a local library for a single music directory
	'''

	def __init__(self, music_dir, taglib):
		self.__music_dir = music_dir
		self.__taglib = taglib

		#Local locks for the builders
		self.__tagLock = Lock()
		self.__trackLock = Lock()
	
		self.complete = False
		
		#Metadata structures
		self.__trackList = {}
		self.__tagList = {}
	
		try:
			self.__mtime = os.stat(music_dir).st_mtime
		except:
			self.__mtime = 0
		self.complete = self.__load_library()
		log.debug('builder prepared for %s' % music_dir)

	def __load_library(self):
		'''
			Fetches library from permanent storage
		'''
		try:
			import json
			fp = open(os.path.join(os.environ['HOME'], '.tagplayer/library_'+str(hash(self.__music_dir))), "r");
			library = json.load(fp)
			fp.close()
			valid = library["music_dir"] == self.__music_dir and library["modified_time"] == self.__mtime
			if valid:
				self.__trackList = library["track"]
				self.__tagList = library["tag"]
			return valid
		except:
			import sys
			log.debug("No stored library found: "+repr(sys.exc_info()))
			return False
	
	def __save_library(self):
		'''
			Stores library to permanent storage
		'''
		import json, time
		library = { 'music_dir': self.__music_dir, 'modified_time': self.__mtime,'track': self.__trackList, 'tag': self.__tagList }
		fp = open(os.path.join(os.environ['HOME'], '.tagplayer/library_'+str(hash(self.__music_dir))), "w");
		json.dump(library, fp , sort_keys=True, indent = 2)
		fp.close()
	
	def get_tracklist(self):
		'''
			Returns the tracklist generated by this Builder instance
		'''
		log.debug('[Builder] get_tracklist {0}'.format(self.__music_dir))
		with self.__trackLock:
			return self.__trackList.copy()
	
	def get_taglist(self):
		'''
			Returns the taglist generated by this Builder instance
		'''
		log.debug('[Builder] get_taglist {0}'.format(self.__music_dir))
		with self.__tagLock:
			return self.__tagList.copy()
			
	def get_playlist(self, tags):
		tgl = self.get_taglist()
		tkl = self.get_tracklist()
		log.debug('builder {0} with {1}-{2}'.format(self.__music_dir, len(tgl), len(tkl)))
		return map(lambda x: tkl[x].__setitem__('path', x) or tkl[x] , Util.flatten(Util.map_to_gen(lambda x: tgl[x]['track_list'], filter(lambda x: x in tags, tgl))))
	
	def build(self):#TODO update through diff instead of full library build on dir change
		'''
			Builds library (including load and update of old library).
		'''
		log.info('[Builder] {0} is complete : {1}'.format(self.__music_dir, self.complete))
		if self.complete:
			return
	
		#generator for all file paths in dir
		path_gen = Util.flatten((os.path.join(r,x) for x in f) for r,_,f in os.walk(self.__music_dir))
		
		#conditions for continuing processing file paths
		ext_cond = lambda x: os.path.splitext(x)[-1] in self.__taglib.music_extensions #has music file extension
		lib_cond = lambda x: x not in self.__trackList #isn't in the library yet
		
		#generator for songs to be processed (added) to library
		new_songs_gen = Util.filter_gen( lambda x : ext_cond(x) and lib_cond(x), path_gen )
		
		#generator for music metadata
		metadata_gen = Util.map_to_gen(lambda y: (y, Util.get_track_metadata(y)), new_songs_gen)

		workers = ThreadPool(10)
		
		workers.map( lambda x : self.__add_to_library(x) , metadata_gen)
		
		self.complete = True
		self.__save_library()
		log.info('[Builder] {0} dir built'.format(self.__music_dir))
	
	#Last.fm information processor
	
	def __add_to_library(self, x):
		'''
			Adds song to library based on last.fm data.
		'''
		try:
			lfm_info = Future(LastFM.get_track_info, x[1][1], x[1][0])
		
			cached = False
		
			with self.__taglib.cacheLock:
				if x[1][0] not in self.__taglib.artistcache:
					self.__taglib.artistcache[x[1][0]] = {'tags': [] , 'lock': Lock()}
					tags = Future(LastFM.get_artist_top_tags, x[1][0])
					self.__taglib.artistcache[x[1][0]]['lock'].acquire()
				else:
					cached = True

			valid = lfm_info() != 'Exception raised within Future' #TODO this treats the symptom, not the disease
			if not valid:
				log.warning("invalid song on last.fm: {} - {} ({})".format(x[1][0], x[1][1],x[0]))
			
			if valid:
				self.__add_track(lfm_info(), x[0])

			if not cached:
				with self.__taglib.cacheLock:
					self.__taglib.artistcache[x[1][0]]['tags']=tags()
					self.__taglib.artistcache[x[1][0]]['lock'].release()
			else:
				with self.__taglib.artistcache[x[1][0]]['lock']:
					tags = lambda: self.__taglib.artistcache[x[1][0]]['tags']
			if valid:
				self.__add_tags(tags(), x[0])
		except:
			import sys
			log.debug('Error on file {0} by {1} - {2}'.format(x[0],x[1][0], sys.exc_info()))

	
	def __add_track(self, lfm_info, path):
		'''
			Adds track's Last.fm info to library's __trackList
		'''
		with self.__trackLock:
			if path not in self.__trackList:
				log.debug('[Builder] adding {0} for {1}'.format(path, self.__music_dir))
				self.__trackList[path] = lfm_info
	
	
	def __add_tags(self, tags, path):
		'''
			Adds tracks' artist tag Last.fm info to library's __tagList
		'''
		with self.__tagLock:
			for tag,weight in tags:
				if weight > 30:
					if tag not in self.__tagList:
						self.__tagList[tag] = {'weight': 0, 'track_list': [] }
					if path not in self.__tagList[tag]['track_list']:
						self.__tagList[tag]['track_list'].append(path)
						self.__tagList[tag]['weight'] += weight

