from __future__ import unicode_literals

from threading import Thread, RLock, Condition
import copy

class ThreadPool:
	'''
		Pool of threads to perform distributed map function.
		Usage is as follows (the same as Pool class from multiprocessing module):
			Create pool by using class constructor, giving as argument the desired number of concurrent threads.
			Call map function of ThreadPool object with desired function and list/generator of objects to be mapped
	'''	
	
	def __init__(self, n_threads=2):
		'''
			Class initialization. Prepares necessary objects to map function.
			Default number of threads is 2 but should be overriden according to need.
		'''
		if n_threads <= 0:
			n_threads = 2
		self.n_threads = n_threads
		
		self.active_threads = 0
		
		self.lock = RLock() # methods: acquire/release
		self.cond = Condition(lock=self.lock) # methods: wait/notify

	def map(self, func, items):
		'''
			Maps function func to each element in items.
			Each execution of func runs in a different thread, allowing for concurrent mapping.
			Order of elements in items is preserved in map result.
		'''	
		gen = (x for x in items)
		result = []
		
		done = False
		
		while(not done):
			
			self.lock.acquire()
			while(self.active_threads >= self.n_threads):
				self.cond.wait()
			self.lock.release()
			
			try:
				item = gen.next()
				
				self.lock.acquire()
#				print 'Locked at index %d' % (len(result))
				result.append(None)
				
				def aux(index, item):
					r = func(item)
					self.lock.acquire()
#					print '\tLock in %d'%(index,)
					result[index] = r
					self.active_threads-=1
#					print '\tUnlock in %d'%(index,)
					self.cond.notifyAll()
					self.lock.release()
				
				t = Thread(target=aux, args=(len(result)-1,item))
			
				t.start()
				
				self.active_threads+=1

#				print 'Unlocked at index %d' % (len(result)-1,)
				self.lock.release()
			
			except StopIteration:
				self.lock.acquire()
				self.cond.notifyAll()
				self.lock.release()
				done = True	
		
		self.lock.acquire()
		while(self.active_threads > 0):
			self.cond.wait()
		self.lock.release()
		
		return result

class Future:
#TODO Credits to this class's creator
    def __init__(self,func,*param):
        # Constructor
        self.__done=0
        self.__result=None
        self.__status='working'

        self.__C=Condition()   # Notify on this Condition when result is ready

        # Run the actual function in a separate thread
        self.__T=Thread(target=self.Wrapper,args=(func,param))
        self.__T.setName("FutureThread")
        self.__T.start()

    def __repr__(self):
        return '<Future at '+hex(id(self))+':'+self.__status+'>'

    def __call__(self):
        self.__C.acquire()
        while self.__done==0:
            self.__C.wait()
        self.__C.release()
        # We deepcopy __result to prevent accidental tampering with it.
        a=copy.deepcopy(self.__result)
        return a

    def Wrapper(self, func, param):
        # Run the actual function, and let us housekeep around it
        self.__C.acquire()
        try:
            self.__result=func(*param)
        except:
            #print repr(param)+repr(func)
            self.__result="Exception raised within Future"
        self.__done=1
        self.__status=`self.__result`
        self.__C.notify()
        self.__C.release()


if __name__ == "__main__":
	'''
		Small test, just to make sure it works
	'''
	test = []
	from random import Random
	r = Random()
	
	for i in range(20000):
		test.append(r.randint(1,100))

	def fact(x):
		if x < 1:
			return 1
		else:
			return x*fact(x-1)
		
	tp = ThreadPool(100)
	import time
	def coiso(x):
		r2=Random()
		time.sleep(r2.randint(1,800)/1000.0)
		return fact(x)
	
	def fail_n(x):
		if len(x) == 0:
			return 0
		else:
			return x[0] is None and 1 or 0 + fail_n(x[1:])
	
	print 'Failed %d' % (fail_n(tp.map(coiso, test)),)
